"""pycmdline.py -- Python/CL command line interface for Midas

Provides this functionality:

- Command directives:
        .exit
        .help
        .complete [ 1 | 0 ]
        .debug
        .mimic
        .info
- Shell escapes (!cmd, !!cmd to run in /bin/sh)
- CL command-mode execution, triggered by a line that starts with a
  CL task token and is not followed by other characters indicating
  it is some other kind of Python statement
- Normal Python mode execution (when CL emulation and directives are
  not triggered)

Uses standard code module plus some ideas from cmd.py module
(and of course Perry's Monty design.)

Original:  R. White, 2000 February 20

Modified for PyMidas usage
"""

##############################################################################
#   This file is part of PyMidas.
#
#   PyMidas is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   PyMidas is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with PyMidas; if not, write to the Free Software
#   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#
#
###############################################################################
#
#  Author:             S. Maisala       
#  Email:              pymidas@eso.org
#  Contact:            www.eso.org/sampo/pymidas
#
##############################################################################
#
#   Change History log
#
# Num  Date      Revision  Who   What
#
# 00   2005/07/15  0.95    SM   Creation of this History  
#
#
# 01   2005/09/22  0.95    SM   Mimic mode change/direc bug corrected
#                               
# 02   2005/09/22  0.95    SM   Help command works also  in  PyMidas mode
#                               commands             
#
# 03   2005/11/03  1.0.0   SM  Version 1.0.0
#
###############################################################################




import string, re, os, sys, code, types, keyword, traceback, linecache
import minmatch, pymidasCom, _pymidasInit, midascompleter, midas




# Init midas tasks
import  midastask
from midas import version

dazUnit = os.environ[ 'DAZUNIT' ]
midversion = os.environ[ 'MIDVERS' ]
parallel = os.environ[ 'PARALLEL' ]

# Read default midas commands
comms = pymidasCom.MidasCommands() 

numMidCom = 0
NumPyMidCom = 0


midasbanner = 'no. Command   (PyMidas v. '+ version+ ', Midas unit '+ dazUnit+ \
              ', version '+ midversion +' ' + parallel+')'

class CmdConsole(code.InteractiveConsole):
    """Base class for command console.

    Similar to InteractiveConsole, but provides local prompt control and
    hook for simple non-Python command processing.
    """

    def __init__(self, locals=None, filename="<console>",
                    cmddict=None, prompt1=">>> ", prompt2="... ",
                    cmdchars=("a-zA-Z_.","0-9",'/')):
        code.InteractiveConsole.__init__(self, locals=locals, filename=filename)
        self.ps1 = prompt1
        self.ps2 = prompt2
        if cmddict==None: cmddict={}
        self.cmddict = cmddict
        # cmdchars gives character set for first character, following
        # characters in the command name
        # create pattern that puts command in group 'cmd' and matches
        # optional leading and trailing whitespace
        self.recmd = re.compile( "[ \t]*(?P<cmd>" +
                "[" + cmdchars[0] + "][" + cmdchars[0] + cmdchars[1] + cmdchars[2] + "]*"  + 
                ")[ \t]*")
        # history is a list of lines entered by user (allocated in blocks)
        self.history = 200*[None]       #PyMidas history
        self.midhistory = 200*[None]    #Midas history
        self.nhistory = 0               #PyMidas num History
        self.midnhistory = 0            #Midas Num History
        print 'Midas '+midversion+' ' + parallel+ '  European Southern Observatory'
        print


    def addHistory(self, line):
        """Append a line to history"""
        if self.nhistory >= len(self.history):
            self.history.extend(100*[None])
        self.history[self.nhistory] = line
        self.nhistory = self.nhistory + 1

    def addMidHistory(self, line):
        """Append a line to history"""
        if self.midnhistory >= len(self.midhistory):
            self.midhistory.extend(100*[None])
        self.midhistory[self.midnhistory] = line
        self.midnhistory = self.midnhistory + 1



    def printHistory(self, n=20):
        """Print last n lines of history"""
        for i in range(-min(n, self.nhistory),0):
            num = min(n, self.nhistory)+1
            print num+i,'   ',self.history[self.nhistory+i]


    def printMidHistory(self, n=20):
        """Print last n lines of history"""
        for i in range(-min(n, self.midnhistory),0):
            num = min(n, self.midnhistory)+1
            print num+i,'   ',self.midhistory[self.midnhistory+i]



    def interact(self, banner=None):
        """Emulate the interactive Python console, with extra commands.

        Also is modified so it does not catch EOFErrors."""
        if banner is None:
            self.write("Python %s on %s\n%s\n(%s)\n" %
                                    (sys.version, sys.platform, sys.copyright,
                                    self.__class__.__name__))
        else:
            self.write("%s\n" % str(banner))
        more = 0
        # number of consecutive EOFs
        neofs = 0
        # flag indicating whether terminal ID needs to be set
        needtermid = 1
        while 1:
            try:
                if not sys.stdin.isatty():
                    prompt = ""
                elif more:
                    prompt = self.ps2
                else:
                    #prompt = self.ps1
                    if pymidasCom.globals.mimic:
                        prompt = 'Midas '+ pymidasCom._modifyNum( self.midnhistory + 1 )+'>'
                    else:
                        prompt = 'PyMidas '+ pymidasCom._modifyNum( self.nhistory + 1 )+'>'
                        
                # reset the focus to terminal if necessary
                #wutil.focusController.resetFocusHistory()
                line = self.raw_input(prompt)
                if needtermid and prompt:
                    # reset terminal window ID immediately
                    # after first input from terminal
                    #wutil.terminal.updateWindowID()
                    needtermid = 0
                neofs = 0
                # add non-null lines to history
                if string.strip(line):
                    # we do not add numbers in history
                    if  not line[:].isdigit():
                        if pymidasCom.globals.mimic:
                            self.addMidHistory(line)                            
                        else:
                            self.addHistory(line)
                            
                        
                # note that this forbids combination of python & CL
                # code -- e.g. a for loop that runs CL tasks.
                if not more:
                    line = self.cmd(line)
                if line or more: more = self.push(line)
            except EOFError:
                #XXX ugly code here -- refers to methods
                #XXX defined in extensions of this class
                neofs = neofs+1
                if neofs>=5:
                    self.write("\nToo many EOFs, exiting now\n")
                    self.do_exit()
                self.write("\nUse .exit to exit\n"
                        ".help describes executive commands\n")
                self.resetbuffer()
                more = 0
            except KeyboardInterrupt:
                self.write("^C\n")
                self.resetbuffer()
                more = 0

    def cmd(self, line):
        """Check for and execute commands from dictionary."""
        mo = self.recmd.match(line)
        if mo is None:
            if line == '':
                print midasbanner
                if pymidasCom.globals.mimic:
                   self.printMidHistory(200) 
                else:
                    self.printHistory(200)
            i = 0
            method_name = None
            cmd =''
            
        else:
            
            cmd = mo.group('cmd')
            i = mo.end()
            # look up command in dictionary
            method_name = self.cmddict.get(cmd)
        if method_name is None:           
            # no method, but have a look at it anyway
            return self.default(cmd,line,i)
        else:
            # if in cmddict, there must be a method by this name
            f = getattr(self, method_name)
            return apply(f, (line, i))

    def default(self, cmd, line, i):
        """Hook to handle other commands (this version does nothing)"""
        return line

# put the executive commands in a minimum match dictionary

_cmdDict = minmatch.QuietMinMatchDict({
                                '.help': 'do_help',
                                '.logfile': 'do_logfile',
                                '.exit': 'do_exit',
                                '.fulltraceback': 'do_fulltraceback',
                                '.complete': 'do_complete',
                                '.debug': 'do_debug',
                                '.xhelp': 'do_xhelp',
                                '.mimic': 'do_mimic',
                                '.info': 'do_info'
                                
                                })

class PyCmdLine(CmdConsole):

    """Simple Python interpreter with executive commands"""

    def __init__(self, locals=None, logfile=None, complete=1, debug=0,
                    clemulate=1):
        CmdConsole.__init__(self, locals=locals,
                cmddict=_cmdDict, prompt1='PyMidas>', prompt2="... ")
        self.reword = re.compile('[a-z]*')
        self.complete = complete
        self.debug = debug
        self.clemulate = clemulate
        self.logfile = None
        self.lasttrace = None
        if logfile is not None:
            if hasattr(logfile,'write'):
                self.logfile = logfile
            elif type(logfile) is types.StringType:
                self.do_logfile(logfile)
            else:
                self.write('logfile ignored -- not string or filehandle\n')
        # turn command completion on or off as requested
        self.do_complete(default=self.complete)

        

    def initTasks(self):
        comms = pymidasCom.MidasCommands()
        for tasks in comms:
            tasks = tasks.lower( )
            tasks = pymidasCom.midas2py(tasks)
            setattr( midas, tasks, midastask.MidasTask(tasks) )
        # Add Pymidas Extra commands
        setattr( midas, 'do', pymidasCom._do ) 
        #setattr( midas, 'getKeyword' , pymidasCom._getKeyVal )
        setattr( midas, 'putKeyword' , pymidasCom._putKeyVal )
        setattr( midas, 'getPix' , pymidasCom._getPix )
        

    def runsource(self, source, filename="<input>", symbol="single"):
        """Compile and run some source in the interpreter.

        Modified from code.py to include logging."""

        # Check do we need variable output
	if 'midas.' in source and '=' not in source:
	    pymidasCom.globals.variable = 0
	else:
            pymidasCom.globals.variable = 1
        try:
            pcode = code.compile_command(source, filename, symbol)
        except (OverflowError, SyntaxError):
            # Case 1
            self.showsyntaxerror(filename)
            return 0

        if pcode is None:
            # Case 2
            return 1

        # Case 3
        
        self.runcode(pcode)
        if self.logfile:
            self.logfile.write(source)
            if source[-1:] != '\n': self.logfile.write('\n')
            self.logfile.flush()
        return 0


    def do_info(self, line='', i=0):
        """Print info on PyMidas commands"""
        if self.debug>1: self.write('do_help: %s\n' % line[i:])
        self.write("""
PyMidas is a extension for the ESO Midas analysis software that was
developed by the Sampo project. 

PyMidas allows the user to run Midas commands under Python with support for
simultaneous usage of other Python-based systems, such as PyRAF. 

>From PyMidas Midas commands are by default invoked using a new syntax where, 
'/' is removed and the normal Midas command extension first letter is made
uppercase. All command extensions have a length of four letters.

As an example the Midas command READ/KEYWORD can be executed from PyMidas
either with text output to the screen with:

PyMidas 001> readKeyw <parameter>

or with the text written into a Python list of text elements with:

PyMidas 002> result = midas.readKeyw(<parameter>)

In addition PyMidas has a "mimic mode" that allows PyMidas usage with 
original Midas syntax. Mimic mode can be switched on with .mimic 1 and
precludes the use of Python features.

For more details please consult documentation available through the Sampo
web pages at: www.eso.org/sampo.

PyMidas is an experimental pilot product and has no formal support
guarantees from ESO.

""")


    def do_help(self, line='', i=0):
        """Print help on executive commands"""
        if self.debug>1: self.write('do_help: %s\n' % line[i:])
        self.write("""

Executive commands (commands can be abbreviated):

.exit
Exit from PyMidas.

.help
Print this help message.

.info
Information  of PyMidas

.complete [0|1]
Turn command-completion on (default) or off.  When on, the tab character
acts as the completion character, attempting to complete a partially
specified task name, variable name, filename, etc.  If the result is
ambiguous, a list of the possibilities is printed.  Use ^V+tab to insert
a tab other than at the line beginning.

.debug [1|0]
Set debugging flag.  If argument is omitted, default is 0 (debugging off.)

.mimic [1|0]
Turn MIDAS mimic on or off (default). When on the MIDAS commands can be given
as the original MIDAS format. This syntax is not working when accessing Python
style midas.command(). 

.xhelp
Open graphical help window for Midas commands

""")

    def do_exit(self, line='', i=0):
        """Exit from Python"""
        if self.debug>1: self.write('do_exit: %s\n' % line[i:])
        pymidasCom._pymidas.Do( 'CLEAR/BACKGROUND 2; BYE' )
        _pymidasInit.cleanSession( self.debug )

        raise SystemExit




    def do_xhelp(self, line='', i=0):
        """Midas graphical help window"""
        if self.debug>1: self.write('do_help: %s\n' % line[i:])
        pymidasCom._pymidas.Do( 'create/gui help' )




    def do_logfile(self, line='', i=0):
        """Start or stop logging commands"""
        if self.debug>1: self.write('do_logfile: %s\n' % line[i:])
        args = string.split(line[i:])
        if len(args) == 0:      # turn off logging (if on)
            if self.logfile:
                self.logfile.close()
                self.logfile = None
            else:
                self.write("No log file currently open\n")
        else:
            filename = args[0]
            del args[0]
            oflag = 'a'
            if len(args) > 0:
                if args[0] == 'overwrite':
                    oflag = 'w'
                    del args[0]
                elif args[0] == 'append':
                    del args[0]
            if args:
                self.write('Ignoring unknown options: %s\n' %
                        string.join(args," "))
            try:
                oldlogfile = self.logfile
                self.logfile = open(filename,oflag)
                if oldlogfile: oldlogfile.close()
            except IOError, e:
                self.write("error opening logfile %s\n%s\n" % (filename,str(e)))
        return ""



    def do_mimic(self, line='', i=0):
        status = 1
        counter = 0
        args = string.split(line[i:])
        if len(args) == 0:
            pymidasCom.globals.mimic = 0
        else:
            pymidasCom.globals.mimic = int(args[0])
            if pymidasCom.globals.mimic:
                print 'MIDAS mimic is on.'
            else:
                print 'MIDAS mimic is off.'                 
                
        return 



    def do_complete(self, line='', i=0, default=1):
        """Turn command completion on or off"""
        if self.debug>1: self.write('do_complete: %s\n' % line[i:])
        self.complete = default

        if line[i:] != "":
            try:
                self.complete = int(line[i:])
            except ValueError, e:
                if self.debug: self.write(str(e)+'\n')
                pass
        if self.complete:
            # set list of executive commands            
            midascompleter.completer.executive(_cmdDict.keys())
            midascompleter.activate()
        else:
            midascompleter.deactivate()
        return ""

    def do_debug(self, line='', i=0):
        """Turn debug output on or off"""
        if self.debug>1: self.write('do_debug: %s\n' % line[i:])
        pymidasCom._pymidas.SetParam ('TRACE_LEVEL',line[i:] )
        pymidasCom._pymidas.SetParam ('TRACE_CONSOLE', line[i:] )
        pymidasCom._pymidas.SetParam ('TRACE_FILE','1')
        self.debug = 1
        pymidasCom.globals.debug = 1
        if line[i:] != "":
            try:
                self.debug = int(line[i:])
                pymidasCom.globals.debug = int( line [ i ] )
            except ValueError, e:
                if self.debug: self.write(str(e)+'\n')
                pass
        return ""

    def do_fulltraceback(self, line='', i=0):
        """Print full version of last traceback"""
        if self.debug>1: self.write('do_fulltraceback: %s\n' % line[i:])
        self.showtraceback(reprint=1)
        return ""

    def default(self, cmd, line, i):
        """Check for MIDAS task calls.

        cmd = alpha-numeric string from beginning of line
        line = full line (including cmd, preceding blanks, etc.)
        i = index in line of first non-blank character following cmd
        """
        
        #print cmd,line,i


        pymidasCom.globals.variable = 0
        if self.debug>1: self.write('default: %s - %s\n' % (cmd,line[i:]))
        if len(cmd)==0:

            if line[i:i+1] == '!':
                # '!' is shell escape
                pymidasCom.clOscmd(line[i+1:])
                return ''                        
                
            
          
            elif (line[i:i+1] =='@' or line[i:i+2] =='@@' or line[i:i+1] =='?') :
                result = pymidasCom._do( line ) 
                #print result
                return 
            
            # Let's try midas command calls by number
            elif  line[i:].isdigit():

                if pymidasCom.globals.mimic:

                    # we have limitation of history commands in midnhistory
                    if ( int( line )-1 < self.midnhistory and self.midnhistory > 0): 
                        if self.midhistory[int(line)-1] =='None':
                        
                            print 'Command no.',line,'not in buffer'
                            return 
                        else:
                            # print self.history[int(line)-1]
                            result = self.cmd(self.midhistory[int(line)-1])
                            return result                         
                    
                    else:
                        print 'Command no. ',line,'not in buffer'
                        return
                else:
                    # we have limitation of history commands in nhistory
                    if ( int( line )-1 < self.nhistory and self.nhistory > 0): 
                        if self.history[int(line)-1] =='None':
                        
                            print 'Command no.',line,'not in buffer'
                            return 
                        else:
                            # print self.history[int(line)-1]
                            result = self.cmd(self.history[int(line)-1])
                            return result                         
                    
                    else:
                        print 'Command no. ',line,'not in buffer'
                        return


                
                
        # Check if MIDAS mimic is on
        elif pymidasCom.globals.mimic:
            testline = line.upper()
            if testline in ['BYE','EXIT']: # If exit command
                self.do_exit(line='', i=0)
	    elif cmd in 'CHANGE/DIRE':   #(01) cd in mimic mode fix
		print line[i+1:]
            	args = line[i:]
            	if args =='':
                	command = 'midas.cd()'
            	else:
                	command = "midas.cd('"+args+"')"
            	return command

            else:
                result =  pymidasCom._run( line )
                # Update command database if necessary
                #print cmd
                #if cmd in ['createComm','CREATE/COMM','setCont','SET/CONT'] :
                #    pymidasCom.readCommands()
                #    self.initTasks()
                # Remove commands from database if necessary    
                #if cmd in ['deleteComm', 'DELETE/COMM']:
                #    pymidasCom.readCommands()
                #    self.initTasks()
                
                return
            
        elif keyword.iskeyword(cmd) or \
          ( os.__builtins__.has_key(cmd) and cmd not in [ 'type', 'dir', 'help' ] ):
            # don't mess with Python keywords or built-in functions
            # except allow 'type', 'dir, 'help' to be used in simple syntax
            return line


        # (02) Here we change Python to Midas syntax in help command
        # This makes possible to use help also in  PyMidas mode commands 
        elif cmd in 'help':
            if pymidasCom.globals.mimic ==0:
                para =  line[i:]
                if  hasattr(midas,para):
                    para = pymidasCom.py2midas(para) 
                    line = cmd + ' '+ para
                    
        # Extra commands to PyMidas
        elif cmd in [ 'cd', 'pwd' ]:
            args = line[i:]
            if args =='':
                command = 'midas.'+ cmd+'()'
            else:    
                command = 'midas.'+ cmd+"('"+args+"')"      
            return command

        #  Extra unix shell commands
        elif cmd in ['ls', 'rm','rmdir', 'cp', 'mv' ]:              
            pymidasCom.clOscmd(line)         
            return ''
        
        elif line[i:i+1] != "" and line[i] in '=,[':
            # don't even try if it doesn't look like a procedure call
            #print 'procedure call'
            pymidasCom.globals.variable = 1
            return line
        elif not hasattr(midas,cmd):
            # not an MIDAS command
            #XXX Eventually want to improve error message for
            #XXX case where user intended to use MIDASsyntax but
            #XXX forgot to load package
            return line
        elif self.isLocal(cmd):
            # cmd is both a local variable and an MIDAS task or procedure name
            # figure out whether MIDAS or CL syntax is intended from syntax
            if line[i:i+1] == "" or line[i] == "(":
                return line
            if line[i] not in string.digits and \
               line[i] not in string.letters and \
               line[i] not in "<>|":
                # this does not look like an MIDAS command
                return line
            # check for some Python operator keywords
            mm = self.reword.match(line[i:])
            if mm.group() in ["is","in","and","or","not"]:
                return line
        elif line[i:i+1] == '(':
            if cmd in ['type', 'dir']:
                # assume a standalone call of Python type, dir functions
                # rather than MIDAS task
                #XXX Use MIDAS help function in every case (may want to
                # change this eventually, when Python built-in help
                # gets a bit better.)

                return line
            else:
                # Not a local function, so user presumably intends to
                # call MIDAS task.  Force Python mode but add the 'midas.'
                # string to the task name for convenience.
                #XXX this find() may be improved with latest Python readline features
                j = string.find(line,cmd)
                #print' midastask',line[:j] + 'midas.' + line[j:]
                return line[:j] + 'midas.' + line[j:]
        elif not callable(getattr(midas,cmd)):
            # variable from midas module is not callable task (e.g.,
            # yes, no, INDEF, etc.) -- add 'midas.' so it can be used
            # as a variable and execute as Python
            j = string.find(line,cmd)
        
            return line[:j] + 'midas.' + line[j:]

        # if we get to here then it looks like MIDAS code
        if self.debug>1: self.write('CL: %s\n' % line)
        try:
             # Check if mimic is on and MIDAS shell command
            if pymidasCom.globals.mimic:                
                if line[i:i+1] =='$':
                    pymidasCom.clOscmd(line[i+1:])
                    return ''

            else:
                result = pymidasCom._run(line)
                # Update command database if necessary
                #if cmd in ['createComm','CREATE/COMM','setCont','SET/CONT'] :
                #    pymidasCom.readCommands()
                #    self.initTasks()
                # Remove commands from database if necessary    
                #if cmd in ['deleteComm', 'DELETE/COMM']:
                #    pymidasCom.readCommands()
                #    self.initTasks()
            #print result
            if self.debug>1: self.write('Midas command succesful: %s\n' % line)
            if self.logfile is not None:
                # log CL code as comment
                cllines = string.split(line,'\n')
                for oneline in cllines:
                    self.logfile.write('# %s\n' % oneline)
                self.logfile.write(code)
                self.logfile.flush()
            return ''
        except:
            self.showtraceback()
        return ''

    def isLocal(self, value):
        """Returns true if value is local variable"""
        ff = string.split(value,'.')
        return self.locals.has_key(ff[0])

    def start(self, banner=
              "*****************************************************************************\n"
              "**                                                                         **\n"
              "**       Copyright (C)2005 SAMPO, European Southern Observatory            **\n"
              "**                                                                         **\n"
              "**   ESO-PyMidas comes with ABSOLUTELY NO WARRANTY; for details type       **\n"
              "**   `@ license w'. This is free software, and you are welcome to          **\n"
              "**   redistribute it under certain conditions; type `@ license c'          **\n"
              "**   for details; .help describes executive commands; .info gives          **\n"
              "**   more information of the PyMidas usage                                 **\n"
              "**                                                                         **\n"
              "*****************************************************************************\n"):
        """PyMidas Banner"""
        self.interact(banner=banner)

    def showtraceback(self, reprint=0):
        """Display the exception that just occurred.

        We remove the first stack item because it is our own code.
        Strip out references to modules within pyraf unless reprint
        or debug is set.
        """
        try:
            if reprint:
                if self.lasttrace is None: return
                type, value, tbmod = self.lasttrace
            else:
                type, value, tb = sys.exc_info()
                linecache.checkcache()
                sys.last_type = type
                sys.last_value = value
                sys.last_traceback = tb
                tblist = traceback.extract_tb(tb)
                del tblist[:1]
                self.lasttrace = type, value, tblist
                if self.debug:
                    tbmod = tblist
                else:
                    tbmod = []
                    for tb1 in tblist:
                        path, filename = os.path.split(tb1[0])
                        path = os.path.normpath(os.path.join(os.getcwd(), path))
            list = traceback.format_list(tbmod)
            if list:
                list.insert(0, "Traceback (innermost last):\n")
            list[len(list):] = traceback.format_exception_only(type, value)
        finally:
            tbmod = tblist = tb = None
        map(self.write, list)
